/*
 * malloc/free by O.Dreesen
 *
 * first TRY:
 *   lists w/magics
 * and now the second TRY
 *   let the kernel map all the stuff (if there is something to do)
 */

#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>

#include <sys/types.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <limits.h>
//#include <sys/shm.h>	/* for PAGE_SIZE */

#ifndef PAGE_SIZE
 #define PAGE_SIZE sysconf (_SC_PAGESIZE)
#endif

#ifndef MREMAP_MAYMOVE
 #define MREMAP_MAYMOVE 1
#endif

/* -- HELPER CODE --------------------------------------------------------- */

#ifndef MAP_FAILED
#define MAP_FAILED ((void*)-1)
#endif

#ifndef NULL
#define NULL ((void*)0)
#endif

typedef struct {
  void*  next;
  size_t size;
} alloc_t;

#define BLOCK_START(b)	(((void*)(b))-sizeof(alloc_t))
#define BLOCK_RET(b)	(((void*)(b))+sizeof(alloc_t))

#define MEM_BLOCK_SIZE	PAGE_SIZE
#define PAGE_ALIGN(s)	(((s)+MEM_BLOCK_SIZE-1)&(unsigned long)(~(MEM_BLOCK_SIZE-1)))

/* a simple mmap :) */
#if defined(i386)
#define REGPARM(x) attribute((regparm(x)))
#else
#define REGPARM(x)
#endif

static void REGPARM(1) *do_mmap(size_t size) {
  return mmap(0, size, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, (size_t)0);
}

/* -- SMALL MEM ----------------------------------------------------------- */

static alloc_t* small_mem[8];

#define SMALL_NR(i)		(MEM_BLOCK_SIZE/(i))

#define MIN_SMALL_SIZE	SMALL_NR(256)		/*   16 /   32 */
#define MAX_SMALL_SIZE	SMALL_NR(2)		/* 2048 / 4096 */

#define GET_SIZE(s)		(MIN_SMALL_SIZE<<get_index((s)))

#define FIRST_SMALL(p)		(((unsigned long)(p))&(~(MEM_BLOCK_SIZE-1)))

static inline int ind_shift() { return (MEM_BLOCK_SIZE==4096)?4:5; }

static size_t REGPARM(1) get_index(size_t _size) {
  register size_t idx=0;
//  if (_size) {	/* we already check this in the callers */
    register size_t size=((_size-1)&(MEM_BLOCK_SIZE-1))>>ind_shift();
    while(size) { size>>=1; ++idx; }
//  }
  return idx;
}

/* small mem */
static void small_free(void*_ptr,size_t _size) REGPARM(2);

static void REGPARM(2) small_free(void*_ptr,size_t _size) {
  alloc_t* ptr=BLOCK_START(_ptr);
  size_t size=_size;
  size_t idx=get_index(size);

#ifdef WANT_FREE_OVERWRITE
  memset(ptr,0x55,size);	/* allways zero out small mem */
#else
#if 0
  memset(ptr,0,size);	/* allways zero out small mem */
#endif
#endif

  ptr->next=small_mem[idx];
  small_mem[idx]=ptr;
}

static void* REGPARM(1) small_malloc(size_t _size) {
  alloc_t *ptr;
  size_t size=_size;
  size_t idx;

  idx=get_index(size);
  ptr=small_mem[idx];

  if (ptr==0)  {	/* no free blocks ? */
    register int i,nr;
    ptr=do_mmap(MEM_BLOCK_SIZE);
    if (ptr==MAP_FAILED) return MAP_FAILED;

    small_mem[idx]=ptr;

    nr=SMALL_NR(size)-1;
    for (i=0;i<nr;i++) {
      ptr->next=(((void*)ptr)+size);
      ptr=ptr->next;
    }
    ptr->next=0;

    ptr=small_mem[idx];
  }

  /* get a free block */
  small_mem[idx]=ptr->next;
  ptr->next=0;

  return ptr;
}

/* -- PUBLIC FUNCTIONS ---------------------------------------------------- */

void tiny_free(void *ptr) {
  register size_t size;
  if (ptr) {
    size=((alloc_t*)BLOCK_START(ptr))->size;
    if (size) {
      if (size<=MAX_SMALL_SIZE)
	small_free(ptr,size);
      else
	munmap(BLOCK_START(ptr),size);
    }
  }
}

void* tiny_malloc(size_t size) {
  alloc_t* ptr;
  size_t need;
  if (!size) goto err_out;
  size+=sizeof(alloc_t);
  if (size<sizeof(alloc_t)) goto err_out;
  if (size<=MAX_SMALL_SIZE) {
    need=GET_SIZE(size);
    ptr=small_malloc(need);
  }
  else {
    need=PAGE_ALIGN(size);
    if (!need) ptr=MAP_FAILED; else ptr=do_mmap(need);
  }
  if (ptr==MAP_FAILED) goto err_out;
  ptr->size=need;
  return BLOCK_RET(ptr);
err_out:
  //(*errno_location())=ENOMEM;
  return 0;
}

#if 0
void* tiny_calloc(size_t nmemb, size_t _size) {
  register size_t size=_size*nmemb;
  if (nmemb && size/nmemb!=_size) {
    //(*errno_location())=ENOMEM;
    return 0;
  }
#ifdef WANT_FREE_OVERWRITE
  if (size<MAX_SMALL_SIZE) {
    void* x=malloc(size);
    memset(x,0,size);
    return x;
  } else
#endif
  return malloc(size);
}
#endif

void* tiny_realloc(void* ptr, size_t _size) {
  register size_t size=_size;
  if (ptr) {
    if (size) {
      alloc_t* tmp=BLOCK_START(ptr);
      size+=sizeof(alloc_t);
      if (size<sizeof(alloc_t)) goto retzero;
      size=(size<=MAX_SMALL_SIZE)?GET_SIZE(size):PAGE_ALIGN(size);
      if (tmp->size!=size) {
	if ((tmp->size<=MAX_SMALL_SIZE)) {
	  void *new=malloc(_size);
	  if (new) {
	    register alloc_t* foo=BLOCK_START(new);
	    size=foo->size;
	    if (size>tmp->size) size=tmp->size;
	    if (size) memcpy(new,ptr,size-sizeof(alloc_t));
	    free(ptr);
	  }
	  ptr=new;
	}
	else {
	  register alloc_t* foo;
	  size=PAGE_ALIGN(size);
	  foo=mremap(tmp,tmp->size,size,MREMAP_MAYMOVE);
	  if (foo==MAP_FAILED) {
retzero:
	    //(*errno_location())=ENOMEM;
	    ptr=0;
	  }
	  else {
	    foo->size=size;
	    ptr=BLOCK_RET(foo);
	  }
	}
      }
    }
    else { /* size==0 */
      free(ptr);
      ptr = NULL;
    }
  }
  else { /* ptr==0 */
    if (size) {
      ptr=malloc(size);
    }
  }
  return ptr;
}

